@@ -185,6 +185,15 @@ pub fn create_plan<VM: VMBinding>(
 /// 3. Initialize all the spaces the plan uses with the heap meta, and the global metadata specs vector.
 /// 4. Create a `SideMetadataSanity` object, and invoke verify_side_metadata_sanity() for each space (or
 ///    invoke verify_side_metadata_sanity() in `CommonPlan`/`BasePlan` for the spaces in the common/base plan).
+///
+/// Methods in this trait:
+///
+/// Only methods that will be overridden by each specific plan should be included in this trait. The trait may
+/// provide a default implementation, and each plan can override the implementation. For methods that won't be
+/// overridden, we should implement those methods in BasePlan (or CommonPlan) and call them from there instead.
+/// We should avoid having methods with the same name in both Plan and BasePlan, as this may confuse people, and
+/// they may call a wrong method by mistake.
+// TODO: Some methods that are not overriden can be moved from the trait to BasePlan.
 pub trait Plan: 'static + Sync + Downcast {
     type VM: VMBinding;
 
@@ -243,6 +252,12 @@ pub trait Plan: 'static + Sync + Downcast {
         self.base().initialized.load(Ordering::SeqCst)
     }
 
+    fn should_trigger_gc_when_heap_is_full(&self) -> bool {
+        self.base()
+            .trigger_gc_when_heap_is_full
+            .load(Ordering::SeqCst)
+    }
+
     fn prepare(&mut self, tls: VMWorkerThread);
     fn release(&mut self, tls: VMWorkerThread);
 
@@ -319,30 +334,31 @@ pub trait Plan: 'static + Sync + Downcast {
         self.get_total_pages() - self.get_pages_used()
     }
 
+    /// The application code has requested a collection.
     fn handle_user_collection_request(&self, tls: VMMutatorThread, force: bool) {
-        if force || !self.options().ignore_system_g_c {
-            info!("User triggering collection");
-            self.base()
-                .user_triggered_collection
-                .store(true, Ordering::Relaxed);
-            self.base().control_collector_context.request();
-            <Self::VM as VMBinding>::VMCollection::block_for_gc(tls);
-        }
+        self.base().handle_user_collection_request(tls, force)
     }
 
-    fn reset_collection_trigger(&self) {
-        self.base()
-            .user_triggered_collection
-            .store(false, Ordering::Relaxed)
+    /// Return whether last GC was an exhaustive attempt to collect the heap.
+    /// For many collectors this is the same as asking whether the last GC was a full heap collection.
+    fn last_collection_was_exhaustive(&self) -> bool {
+        self.last_collection_full_heap()
     }
 
+    /// Return whether last GC is a full GC.
+    fn last_collection_full_heap(&self) -> bool {
+        true
+    }
+
+    /// Force the next collection to be full heap.
+    fn force_full_heap_collection(&self) {}
+
     fn modify_check(&self, object: ObjectReference) {
-        if self.base().gc_in_progress_proper() && object.is_movable() {
-            panic!(
-                "GC modifying a potentially moving object via Java (i.e. not magic) obj= {}",
-                object
-            );
-        }
+        assert!(
+            !(self.base().gc_in_progress_proper() && object.is_movable()),
+            "GC modifying a potentially moving object via Java (i.e. not magic) obj= {}",
+            object
+        );
     }
 }
 
@@ -359,13 +375,18 @@ pub enum GcStatus {
 BasePlan should contain all plan-related state and functions that are _fundamental_ to _all_ plans.  These include VM-specific (but not plan-specific) features such as a code space or vm space, which are fundamental to all plans for a given VM.  Features that are common to _many_ (but not intrinsically _all_) plans should instead be included in CommonPlan.
 */
 pub struct BasePlan<VM: VMBinding> {
-    // Whether MMTk is now ready for collection. This is set to true when enable_collection() is called.
+    /// Whether MMTk is now ready for collection. This is set to true when initialize_collection() is called.
     pub initialized: AtomicBool,
+    /// Should we trigger a GC when the heap is full? It seems this should always be true. However, we allow
+    /// bindings to temporarily disable GC, at which point, we do not trigger GC even if the heap is full.
+    pub trigger_gc_when_heap_is_full: AtomicBool,
     pub gc_status: Mutex<GcStatus>,
     pub last_stress_pages: AtomicUsize,
     pub stacks_prepared: AtomicBool,
     pub emergency_collection: AtomicBool,
     pub user_triggered_collection: AtomicBool,
+    pub internal_triggered_collection: AtomicBool,
+    pub last_internal_triggered_collection: AtomicBool,
     // Has an allocation succeeded since the emergency collection?
     pub allocation_success: AtomicBool,
     // Maximum number of failed attempts by a single thread
@@ -384,7 +405,7 @@ pub struct BasePlan<VM: VMBinding> {
     pub scanned_stacks: AtomicUsize,
     pub mutator_iterator_lock: Mutex<()>,
     // A counter that keeps tracks of the number of bytes allocated since last stress test
-    pub allocation_bytes: AtomicUsize,
+    allocation_bytes: AtomicUsize,
     // Wrapper around analysis counters
     #[cfg(feature = "analysis")]
     pub analysis_manager: AnalysisManager<VM>,
@@ -490,11 +511,14 @@ impl<VM: VMBinding> BasePlan<VM> {
             ),
 
             initialized: AtomicBool::new(false),
+            trigger_gc_when_heap_is_full: AtomicBool::new(true),
             gc_status: Mutex::new(GcStatus::NotInGC),
             last_stress_pages: AtomicUsize::new(0),
             stacks_prepared: AtomicBool::new(false),
             emergency_collection: AtomicBool::new(false),
             user_triggered_collection: AtomicBool::new(false),
+            internal_triggered_collection: AtomicBool::new(false),
+            last_internal_triggered_collection: AtomicBool::new(false),
             allocation_success: AtomicBool::new(false),
             max_collection_attempts: AtomicUsize::new(0),
             cur_collection_attempts: AtomicUsize::new(0),
@@ -543,6 +567,40 @@ impl<VM: VMBinding> BasePlan<VM> {
         }
     }
 
+    /// The application code has requested a collection.
+    pub fn handle_user_collection_request(&self, tls: VMMutatorThread, force: bool) {
+        if force || !self.options.ignore_system_g_c {
+            info!("User triggering collection");
+            self.user_triggered_collection
+                .store(true, Ordering::Relaxed);
+            self.control_collector_context.request();
+            VM::VMCollection::block_for_gc(tls);
+        }
+    }
+
+    /// MMTK has requested stop-the-world activity (e.g., stw within a concurrent gc).
+    // This is not used, as we do not have a concurrent plan.
+    #[allow(unused)]
+    pub fn trigger_internal_collection_request(&self) {
+        self.last_internal_triggered_collection
+            .store(true, Ordering::Relaxed);
+        self.internal_triggered_collection
+            .store(true, Ordering::Relaxed);
+        self.control_collector_context.request();
+    }
+
+    /// Reset collection state information.
+    pub fn reset_collection_trigger(&self) {
+        self.last_internal_triggered_collection.store(
+            self.internal_triggered_collection.load(Ordering::SeqCst),
+            Ordering::Relaxed,
+        );
+        self.internal_triggered_collection
+            .store(false, Ordering::SeqCst);
+        self.user_triggered_collection
+            .store(false, Ordering::Relaxed);
+    }
+
     // Depends on what base spaces we use, unsync may be unused.
     pub fn get_pages_used(&self) -> usize {
         // Depends on what base spaces we use, pages may be unchanged.
@@ -617,7 +675,7 @@ impl<VM: VMBinding> BasePlan<VM> {
         self.vm_space.release();
     }
 
-    pub fn set_collection_kind(&self) {
+    pub fn set_collection_kind<P: Plan>(&self, plan: &P) {
         self.cur_collection_attempts.store(
             if self.is_user_triggered_collection() {
                 1
@@ -628,13 +686,13 @@ impl<VM: VMBinding> BasePlan<VM> {
         );
 
         let emergency_collection = !self.is_internal_triggered_collection()
-            && self.last_collection_was_exhaustive()
+            && plan.last_collection_was_exhaustive()
             && self.cur_collection_attempts.load(Ordering::Relaxed) > 1;
         self.emergency_collection
             .store(emergency_collection, Ordering::Relaxed);
 
         if emergency_collection {
-            self.force_full_heap_collection();
+            plan.force_full_heap_collection();
         }
     }
 
@@ -666,10 +724,6 @@ impl<VM: VMBinding> BasePlan<VM> {
         *self.gc_status.lock().unwrap() == GcStatus::GcProper
     }
 
-    pub fn is_user_triggered_collection(&self) -> bool {
-        self.user_triggered_collection.load(Ordering::Relaxed)
-    }
-
     fn determine_collection_attempts(&self) -> usize {
         if !self.allocation_success.load(Ordering::Relaxed) {
             self.max_collection_attempts.fetch_add(1, Ordering::Relaxed);
@@ -681,18 +735,26 @@ impl<VM: VMBinding> BasePlan<VM> {
         self.max_collection_attempts.load(Ordering::Relaxed)
     }
 
-    fn is_internal_triggered_collection(&self) -> bool {
-        // FIXME
-        false
+    /// Return true if this collection was triggered by application code.
+    pub fn is_user_triggered_collection(&self) -> bool {
+        self.user_triggered_collection.load(Ordering::Relaxed)
     }
 
-    fn last_collection_was_exhaustive(&self) -> bool {
-        true
+    /// Return true if this collection was triggered internally.
+    pub fn is_internal_triggered_collection(&self) -> bool {
+        let is_internal_triggered = self
+            .last_internal_triggered_collection
+            .load(Ordering::SeqCst);
+        // Remove this assertion when we have concurrent GC.
+        assert!(
+            !is_internal_triggered,
+            "We have no concurrent GC implemented. We should not have internally triggered GC"
+        );
+        is_internal_triggered
     }
 
-    fn force_full_heap_collection(&self) {}
-
-    pub fn increase_allocation_bytes_by(&self, size: usize) {
+    /// Increase the allocation bytes and return the current allocation bytes after increasing
+    pub fn increase_allocation_bytes_by(&self, size: usize) -> usize {
         let old_allocation_bytes = self.allocation_bytes.fetch_add(size, Ordering::SeqCst);
         trace!(
             "Stress GC: old_allocation_bytes = {}, size = {}, allocation_bytes = {}",
@@ -700,25 +762,22 @@ impl<VM: VMBinding> BasePlan<VM> {
             size,
             self.allocation_bytes.load(Ordering::Relaxed),
         );
+        old_allocation_bytes + size
     }
 
-    #[inline]
-    pub(super) fn stress_test_gc_required(&self) -> bool {
-        let stress_factor = self.options.stress_factor;
-        if self.initialized.load(Ordering::SeqCst)
-            && (self.allocation_bytes.load(Ordering::SeqCst) > stress_factor)
-        {
-            trace!(
-                "Stress GC: allocation_bytes = {}, stress_factor = {}",
-                self.allocation_bytes.load(Ordering::Relaxed),
-                stress_factor
-            );
-            trace!("Doing stress GC");
-            self.allocation_bytes.store(0, Ordering::SeqCst);
-            true
-        } else {
-            false
-        }
+    /// Check if the options are set for stress GC. If either stress_factor or analysis_factor is set,
+    /// we should do stress GC.
+    pub fn is_stress_test_gc_enabled(&self) -> bool {
+        use crate::util::constants::DEFAULT_STRESS_FACTOR;
+        self.options.stress_factor != DEFAULT_STRESS_FACTOR
+            || self.options.analysis_factor != DEFAULT_STRESS_FACTOR
+    }
+
+    /// Check if we should do a stress GC now. If GC is initialized and the allocation bytes exceeds
+    /// the stress factor, we should do a stress GC.
+    pub fn should_do_stress_gc(&self) -> bool {
+        self.initialized.load(Ordering::SeqCst)
+            && (self.allocation_bytes.load(Ordering::SeqCst) > self.options.stress_factor)
     }
 
     pub(super) fn collection_required<P: Plan>(
@@ -727,7 +786,17 @@ impl<VM: VMBinding> BasePlan<VM> {
         space_full: bool,
         _space: &dyn Space<VM>,
     ) -> bool {
-        let stress_force_gc = self.stress_test_gc_required();
+        let stress_force_gc = self.should_do_stress_gc();
+        if stress_force_gc {
+            debug!(
+                "Stress GC: allocation_bytes = {}, stress_factor = {}",
+                self.allocation_bytes.load(Ordering::Relaxed),
+                self.options.stress_factor
+            );
+            debug!("Doing stress GC");
+            self.allocation_bytes.store(0, Ordering::SeqCst);
+        }
+
         debug!(
             "self.get_pages_reserved()={}, self.get_total_pages()={}",
             plan.get_pages_reserved(),
@@ -849,12 +918,46 @@ impl<VM: VMBinding> CommonPlan<VM> {
         self.base.release(tls, full_heap)
     }
 
-    pub fn schedule_common<E: ProcessEdgesWork<VM = VM>>(
+    /// Schedule all the common work packets
+    pub fn schedule_common<
+        P: Plan<VM = VM>,
+        E: ProcessEdgesWork<VM = VM>,
+        C: CopyContext<VM = VM> + GCWorkerLocal,
+    >(
         &self,
+        plan: &'static P,
         constraints: &'static PlanConstraints,
         scheduler: &GCWorkScheduler<VM>,
     ) {
-        // Schedule finalization
+        use crate::scheduler::gc_work::*;
+
+        // Stop & scan mutators (mutator scanning can happen before STW)
+        scheduler.work_buckets[WorkBucketStage::Unconstrained].add(StopMutators::<E>::new());
+
+        // Prepare global/collectors/mutators
+        scheduler.work_buckets[WorkBucketStage::Prepare].add(Prepare::<P, C>::new(plan));
+
+        // VM-specific weak ref processing
+        scheduler.work_buckets[WorkBucketStage::RefClosure].add(ProcessWeakRefs::<E>::new());
+
+        // Release global/collectors/mutators
+        scheduler.work_buckets[WorkBucketStage::Release].add(Release::<P, C>::new(plan));
+
+        // Analysis GC work
+        #[cfg(feature = "analysis")]
+        {
+            use crate::util::analysis::GcHookWork;
+            scheduler.work_buckets[WorkBucketStage::Unconstrained].add(GcHookWork);
+        }
+
+        // Sanity
+        #[cfg(feature = "sanity")]
+        {
+            use crate::util::sanity::sanity_checker::ScheduleSanityGC;
+            scheduler.work_buckets[WorkBucketStage::Final].add(ScheduleSanityGC::<P, C>::new(plan));
+        }
+
+        // Finalization
         if !self.base.options.no_finalizer {
             use crate::util::finalizable_processor::{Finalization, ForwardFinalization};
             // finalization
@@ -865,6 +968,9 @@ impl<VM: VMBinding> CommonPlan<VM> {
                     .add(ForwardFinalization::<E>::new());
             }
         }
+
+        // Set EndOfGC to run at the end
+        scheduler.set_finalizer(Some(EndOfGC));
     }
 
     pub fn stacks_prepared(&self) -> bool {
