@@ -69,7 +69,8 @@ impl<VM: VMBinding> SFT for LargeObjectSpace<VM> {
             Some(Ordering::SeqCst),
         );
 
-        if self.common.needs_log_bit {
+        // If this object is freshly allocated, we do not set it as unlogged
+        if !alloc && self.common.needs_log_bit {
             VM::VMObjectModel::GLOBAL_LOG_BIT_SPEC.mark_as_unlogged::<VM>(object, Ordering::SeqCst);
         }
 
@@ -187,6 +188,11 @@ impl<VM: VMBinding> LargeObjectSpace<VM> {
                 let cell = VM::VMObjectModel::object_start_ref(object);
                 self.treadmill.copy(cell, nursery_object);
                 self.clear_nursery(object);
+                // We just moved the object out of the logical nursery, mark it as unlogged.
+                if nursery_object && self.common.needs_log_bit {
+                    VM::VMObjectModel::GLOBAL_LOG_BIT_SPEC
+                        .mark_as_unlogged::<VM>(object, Ordering::SeqCst);
+                }
                 trace.process_node(object);
             }
         }
