@@ -14,6 +14,9 @@ pub use large_object_allocator::LargeObjectAllocator;
 mod malloc_allocator;
 pub use malloc_allocator::MallocAllocator;
 
+pub mod immix_allocator;
+pub use self::immix_allocator::ImmixAllocator;
+
 pub(crate) mod dump_linear_scan;
 pub(crate) mod embedded_meta_data;
 pub(crate) mod linear_scan;
