@@ -6,6 +6,15 @@ use crate::vm::VMBinding;
 
 /// VM-specific methods for garbage collection.
 pub trait Collection<VM: VMBinding> {
+    /// If true, only the coordinator thread can call stop_all_mutators and the resume_mutators methods.
+    /// If false, any GC thread can call these methods.
+    ///
+    /// This constant exists because some VMs require the thread that resumes a thread to be the same thread that
+    /// stopped it.  The MMTk Core will use the appropriate thread to stop or start the world according to the value of
+    /// this constant.  If a VM does not have such a requirement, the VM binding shall set this to false to reduce an
+    /// unnecessary context switch.
+    const COORDINATOR_ONLY_STW: bool = true;
+
     /// Stop all the mutator threads. MMTk calls this method when it requires all the mutator to yield for a GC.
     /// This method is called by a single thread in MMTk (the GC controller).
     /// This method should not return until all the threads are yielded.
@@ -33,11 +42,11 @@ pub trait Collection<VM: VMBinding> {
     /// Ask the VM to spawn a GC thread for MMTk. A GC thread may later call into the VM through these VM traits. Some VMs
     /// have assumptions that those calls needs to be within VM internal threads.
     /// As a result, MMTk does not spawn GC threads itself to avoid breaking this kind of assumptions.
-    /// MMTk calls this method to spawn GC threads during [`enable_collection()`](../memory_manager/fn.enable_collection.html).
+    /// MMTk calls this method to spawn GC threads during [`initialize_collection()`](../memory_manager/fn.initialize_collection.html).
     ///
     /// Arguments:
     /// * `tls`: The thread pointer for the parent thread that we spawn new threads from. This is the same `tls` when the VM
-    ///   calls `enable_collection()` and passes as an argument.
+    ///   calls `initialize_collection()` and passes as an argument.
     /// * `ctx`: The GC worker context for the GC thread. If `None` is passed, it means spawning a GC thread for the GC controller,
     ///   which does not have a worker context.
     fn spawn_worker_thread(tls: VMThread, ctx: Option<&GCWorker<VM>>);
